package be.iminds.ilabt.jfed.espec.bundle;

import be.iminds.ilabt.jfed.espec.model.ExperimentSpecification;
import be.iminds.ilabt.jfed.espec.parser.ExperimentSpecificationParser;
import be.iminds.ilabt.jfed.util.IOUtils;
import org.apache.commons.compress.archivers.ArchiveEntry;
import org.apache.commons.compress.archivers.ArchiveException;
import org.apache.commons.compress.archivers.ArchiveInputStream;
import org.apache.commons.compress.archivers.ArchiveStreamFactory;
import org.apache.commons.compress.compressors.CompressorException;
import org.apache.commons.compress.compressors.CompressorInputStream;
import org.apache.commons.compress.compressors.CompressorStreamFactory;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import java.io.*;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Objects;
import java.util.stream.Collectors;

/**
 * Bundle
 */
public class LocalArchiveESpecBundle implements ESpecBundle {
    private static final Logger LOG = LoggerFactory.getLogger(LocalArchiveESpecBundle.class);

    @Nonnull private final File archiveFile;
    @Nonnull private final List<ArchiveEntry> archiveEntries;
    @Nonnull private final String eSpecYmlContent;

    private final boolean isCompressedArchive;

    @Nonnull private final CompressorStreamFactory compressorStreamFactory;
    @Nonnull private final ArchiveStreamFactory archiveStreamFactory;


    @Nonnull private final ExperimentSpecification eSpec;

    LocalArchiveESpecBundle(@Nonnull File archiveFile) throws ESpecBundleInitException{
        this.archiveFile = archiveFile;

        this.archiveEntries = new ArrayList<>();
        String foundESpecYmlContent = null;

//        boolean isCompressedArchive = archiveFile.getName().endsWith(".tar.gz") || archiveFile.getName().endsWith(".tar.bz2") || archiveFile.getName().endsWith(".tgz");

        compressorStreamFactory = new CompressorStreamFactory();
        archiveStreamFactory = new ArchiveStreamFactory();

        //peek to check if this is a compressed archive, like .tar.gz or .tar.bz2 or .tgz
        boolean isCompressedArchive;
        try (FileInputStream fileInputStream = new FileInputStream(archiveFile);
             BufferedInputStream bufferedInputStream = new BufferedInputStream(fileInputStream);
             CompressorInputStream cis = compressorStreamFactory.createCompressorInputStream(bufferedInputStream)
        ) {
            isCompressedArchive = true;
        } catch (CompressorException | IOException e) {
            isCompressedArchive = false;
        }
        this.isCompressedArchive = isCompressedArchive;
        LOG.debug("Archive "+archiveFile.getPath()+" is "+(isCompressedArchive?"a":"NOT a")+" compressed archive");

        try (FileInputStream fileInputStream = new FileInputStream(archiveFile);
             BufferedInputStream bufferedInputStream = new BufferedInputStream(fileInputStream);
             ArchiveInputStream archiveInput = isCompressedArchive ?
                     archiveStreamFactory.createArchiveInputStream(
                             new BufferedInputStream(compressorStreamFactory.createCompressorInputStream(bufferedInputStream))
                     ) :
                     archiveStreamFactory.createArchiveInputStream(bufferedInputStream)
        ) {
            ArchiveEntry archiveEntry = archiveInput.getNextEntry();
            while (archiveEntry != null) {
                archiveEntries.add(archiveEntry);

                if (archiveEntry.getName().equals(ESPEC_DESCRIPTION_FILENAME)) {
                    byte[] content = new byte[(int) archiveEntry.getSize()];
                    int offset = 0;
                    while(offset < archiveEntry.getSize()) {
                        int read = archiveInput.read(content, offset, content.length - offset);
                        if (read == -1)
                            throw new IOException("Unexpected end of stream");
                        offset += read;
                        assert offset <= archiveEntry.getSize() : "Read to much from stream. This is unexpected. " +
                                "read="+read+" offset="+offset+" archiveEntry.getSize()="+archiveEntry.getSize();
                    }
                    foundESpecYmlContent = new String(content, StandardCharsets.UTF_8);
                }

                archiveEntry = archiveInput.getNextEntry();
            }
            archiveInput.close();
        } catch (IOException | CompressorException | ArchiveException e) {
            throw new ESpecBundleInitException("Failed to process ESpec archive", e);
        }

        if (foundESpecYmlContent != null) {
            this.eSpecYmlContent = foundESpecYmlContent;
        } else {
            throw new RuntimeException("Archive \"" + archiveFile.getPath() + "\" does not contain the required file \""+ ESPEC_DESCRIPTION_FILENAME +"\"");
        }

        ExperimentSpecificationParser parser = new ExperimentSpecificationParser();
        try {
            eSpec = parser.parse(this.getExperimentSpecificationYml());
        } catch (ExperimentSpecificationParser.ExperimentSpecificationParseException e) {
            throw new IllegalArgumentException("Failed to parse ExperimentSpecification", e);
        } catch (IOException e) {
            throw new IllegalArgumentException("Failed to read ExperimentSpecification", e);
        }
    }

    @Nonnull
    @Override
    public List<String> getFileNames() {
        return archiveEntries.stream().map(ArchiveEntry::getName).collect(Collectors.toList());
    }

    @Nonnull
    public byte[] getFileContent(@Nonnull String fileName) throws IOException {

        try (FileInputStream fileInputStream = new FileInputStream(archiveFile);
             BufferedInputStream bufferedInputStream = new BufferedInputStream(fileInputStream);
             ArchiveInputStream archiveInput = isCompressedArchive ?
                     archiveStreamFactory.createArchiveInputStream(new BufferedInputStream(compressorStreamFactory.createCompressorInputStream(bufferedInputStream))) :
                     archiveStreamFactory.createArchiveInputStream(bufferedInputStream)
        ) {
            ArchiveEntry archiveEntry = archiveInput.getNextEntry();
            while (archiveEntry != null) {
                archiveEntries.add(archiveEntry);

                if (archiveEntry.getName().equals(fileName)) {
                    byte[] content = new byte[(int) archiveEntry.getSize()];
                    int offset = 0;
                    while(offset < archiveEntry.getSize()) {
                        int read = archiveInput.read(content, offset, content.length - offset);
                        if (read == -1)
                            throw new IOException("Unexpected end of stream");
                        offset += read;
                        assert offset <= archiveEntry.getSize() : "Read to much from stream. This is unexpected. " +
                                "read="+read+" offset="+offset+" archiveEntry.getSize()="+archiveEntry.getSize();
                    }
                    return content;
                }

                archiveEntry = archiveInput.getNextEntry();
            }
            archiveInput.close();
        } catch (ArchiveException | CompressorException e) {
            throw new IOException("Failed to process ESpec archive", e);
        }

        throw new IllegalArgumentException("There is no file with name \""+fileName+"\"");
    }

    @Override
    public boolean isDir(@Nonnull String fileName) throws IOException {
        return archiveEntries.stream()
                .filter(Objects::nonNull)
                .filter(e -> Objects.equals(e.getName(), fileName))
                .findAny()
                .orElseThrow(() -> new IllegalArgumentException("There is no file with name \""+fileName+"\""))
                .isDirectory();
    }

    @Nonnull
    @Override
    public ExperimentSpecification getExperimentSpecification() {
        return eSpec;
    }
}
